# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/dreamai_pdf',
                'doc_host': 'https://HamzaFarhan.github.io',
                'git_url': 'https://github.com/HamzaFarhan/dreamai_pdf',
                'lib_path': 'dreamai_pdf'},
  'syms': { 'dreamai_pdf.core': { 'dreamai_pdf.core.cid_to_char': ('core.html#cid_to_char', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.color_fill_pdf_text': ('core.html#color_fill_pdf_text', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.combine_lines': ('core.html#combine_lines', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.combine_splits': ('core.html#combine_splits', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.get_avg_gap': ('core.html#get_avg_gap', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.get_max_gap': ('core.html#get_max_gap', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.load_pdf': ('core.html#load_pdf', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.only_bnw': ('core.html#only_bnw', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.pdf_img_to_np': ('core.html#pdf_img_to_np', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.process_text': ('core.html#process_text', 'dreamai_pdf/core.py'),
                                  'dreamai_pdf.core.split_words': ('core.html#split_words', 'dreamai_pdf/core.py')},
            'dreamai_pdf.imports': {},
            'dreamai_pdf.parse': { 'dreamai_pdf.parse.AdaptiveConcatPool2d': ('parse.html#adaptiveconcatpool2d', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.AdaptiveConcatPool2d.__init__': ( 'parse.html#adaptiveconcatpool2d.__init__',
                                                                                        'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.AdaptiveConcatPool2d.forward': ( 'parse.html#adaptiveconcatpool2d.forward',
                                                                                       'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.ColumnCounter': ('parse.html#columncounter', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.ColumnCounter.draw': ('parse.html#columncounter.draw', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.Flatten': ('parse.html#flatten', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.Flatten.forward': ('parse.html#flatten.forward', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.HeadModel': ('parse.html#headmodel', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.HeadModel.__init__': ('parse.html#headmodel.__init__', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.HeadModel.forward': ('parse.html#headmodel.forward', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.LinBnDrop': ('parse.html#linbndrop', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.LinBnDrop.__init__': ('parse.html#linbndrop.__init__', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.col_clusters': ('parse.html#col_clusters', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.create_head': ('parse.html#create_head', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.create_model': ('parse.html#create_model', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.get_n_cols': ('parse.html#get_n_cols', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.load_cols_model': ('parse.html#load_cols_model', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.pdf_cols_to_text': ('parse.html#pdf_cols_to_text', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.pdf_to_batch': ('parse.html#pdf_to_batch', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.pdf_to_cols': ('parse.html#pdf_to_cols', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.pdf_to_text': ('parse.html#pdf_to_text', 'dreamai_pdf/parse.py'),
                                   'dreamai_pdf.parse.pred_cols': ('parse.html#pred_cols', 'dreamai_pdf/parse.py')},
            'dreamai_pdf.segment': { 'dreamai_pdf.segment.get_contact_dict': ('segment.html#get_contact_dict', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.get_edu_dicts': ('segment.html#get_edu_dicts', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.get_job_dicts': ('segment.html#get_job_dicts', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.load_ner_model': ('segment.html#load_ner_model', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.load_segs_model': ('segment.html#load_segs_model', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.ners_to_dicts': ('segment.html#ners_to_dicts', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.segment_to_ners': ('segment.html#segment_to_ners', 'dreamai_pdf/segment.py'),
                                     'dreamai_pdf.segment.text_to_segments': ('segment.html#text_to_segments', 'dreamai_pdf/segment.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_segment.ipynb.

# %% auto 0
__all__ = ['text_to_segments', 'segment_to_ners', 'ners_to_dicts', 'get_edu_dicts', 'get_job_dicts', 'get_contact_dict',
           'load_segs_model', 'load_ner_model']

# %% ../nbs/02_segment.ipynb 3
from .core import *
from .parse import *
from .imports import *


# %% ../nbs/02_segment.ipynb 4
def text_to_segments(text, labeling_model, tags=['education', 'work experience']):
    segs = defaultdict(list)
    for txt in text:
        pred = tags[labeling_model(txt, tags)[0][0]]
        segs[pred].append(txt)
    return segs

def segment_to_ners(text, tagger):
    if is_list(text):
        text = ' '.join(text)
    s = Sentence(text)
    tagger.predict(s)
    return s

def ners_to_dicts(s, search_tags=['ORG', 'DATE'], dict_keys=['COMPANY', 'DATE']):
    tags_list = []
    tags_dict = {}
    for l in s.labels:
        dp = l.data_point
        tag = dp.tag
        for s,k in zip(search_tags, dict_keys):
            if tag == s:
                if not tags_dict.get(k,None):
                    tags_dict[k] = dp.text.strip()
                else:
                    tags_list.append(tags_dict)
                    tags_dict = {k:dp.text.strip()}
                
    return tags_list

def get_edu_dicts(edu, tagger):
    edu = segment_to_ners(edu, tagger)
    edu_list = ners_to_dicts(edu, search_tags=['ORG', 'DATE'], dict_keys=['INSTITUTE', 'DATE'])
    edu_list = [d for d in edu_list if d.get('INSTITUTE', None) is not None]
    return edu_list

def get_job_dicts(job, tagger):
    job = segment_to_ners(job, tagger)
    job_dict = ners_to_dicts(job, search_tags=['ORG', 'DATE'], dict_keys=['COMPANY', 'DATE'])
    job_dict = [d for d in job_dict if d.get('COMPANY', None) is not None]
    return job_dict

def get_contact_dict(text):
    if is_list(text): text = ' '.join(text)
    mail_regex = re.compile(r'[\w.+-]+@[\w-]+\.[\w.-]+')
    phone_regex = re.compile(r'[\d]{3}[\s-]?[\d]{3}[\s-]?[\d]{4}')
    emails = re.findall(mail_regex, text.lower())
    phones = re.findall(phone_regex, text.lower())
    return {'EMAIL':emails, 'PHONE':phones}

def load_segs_model():
    return Labels("roberta-large-mnli")

def load_ner_model():
    return Classifier.load('ner-ontonotes-large')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['load_pdf', 'pdf_img_to_np', 'cid_to_char', 'combine_lines', 'get_avg_gap', 'get_max_gap', 'combine_splits',
           'split_words', 'process_text', 'only_bnw', 'color_fill_pdf_text']

# %% ../nbs/00_core.ipynb 3
from .imports import *

# %% ../nbs/00_core.ipynb 4
def load_pdf(pdf):
    if path_or_str(pdf):
        pdf = pdfplumber.open(pdf)
    elif not isinstance(pdf, pdfplumber.pdf.PDF):
        raise Exception(f'pdf must be a path or pdfplumber.pdf.PDF object.')
    return pdf

def pdf_img_to_np(img):
    return np.array(img.annotated)

def cid_to_char(cidx):
    try:
        return chr(int(re.findall(r'\(cid\:(\d+)\)',cidx)[0]) + 29)
    except:
        return cidx

def combine_lines(txt):
    avg_len = np.mean([len(w.split()) for w in txt])
    txt2 = []
    for i,w in enumerate(txt):
        w = process_text(w)
        if i==0:
            txt2.append(w)
        else:
            if len(txt2[-1].split()) < avg_len:
                txt2[-1]+=' '+w
            else:
                txt2.append(w)
    return txt2

def get_avg_gap(words, key0='top', key1='bottom'):
    if key1 is None: key1 = key0
    return np.mean([w[key0]-words[i-1][key1] for i,w in enumerate(words) if i>0])

def get_max_gap(words, key0='top', key1='bottom'):
    if key1 is None: key1 = key0
    return np.max([w[key0]-words[i-1][key1] for i,w in enumerate(words) if i>0])

def combine_splits(splits):
    avg_len = np.mean([len(s) for s in splits])
    splits2 = []
    for i,s in enumerate(splits):
        if i==0:
            splits2.append(s)
        else:
            if len(splits2[-1]) < avg_len:
                splits2[-1]+=s
            else:
                splits2.append(s)
    return splits2

def split_words(words, key0='top', key1='bottom', avg_gap=None, fill_empty=False):
    if key1 is None: key1 = key0
    if avg_gap is None:
        avg_gap = np.mean([w[key0]-words[i-1][key1] for i,w in enumerate(words) if i>0])
    splits = []
    for i,w in enumerate(words):
        if i==0:
            splits.append([w])
        else:
            if w[key0]-words[i-1][key1] > avg_gap:
                if fill_empty:
                    splits.append(['*']*len(splits[-1]) + [w])
                else:
                    splits.append([w])
            else:
                splits[-1].append(w)
    return splits

def process_text(text):
    text = cid_to_char(text)
    text = re.sub(r"\uf0b7", " ", text)
    text = re.sub(r"\(cid:\d{0,3}\)", " ", text)
    text = re.sub(r'• ', " ", text)
    text = re.sub(r'● ', " ", text)
    return text

def only_bnw(img, thresh=50):
    black = np.all(img < thresh, axis=-1)
    img[~black] = 255
    return img

def color_fill_pdf_text(pdf, height=512, color='black'):
    """
    Gets a pdf and returns a list of images with the text filled in with the specified color.
    
    Parameters
    ----------
    pdf: str or Path or pdfplumber.pdf.PDF
        The path to the pdf or a pdfplumber.pdf.PDF object.
    height: int
        The height of the image to be returned.
    color: str or tuple
        The color to fill the text with.
        
    Returns
    -------
    list
        A list of np.ndarray images.
    """
    
    pdf = load_pdf(pdf)
    color = tuple(color_to_rgb(color))
    return [only_bnw(pdf_img_to_np(page.to_image(height=height).draw_rects(page.extract_words(x_tolerance=5),
                                                                  fill=color, stroke=color))) for page in pdf.pages]



import pdfplumber
from flair.nn import Classifier
from flair.data import Sentence
from sklearn.cluster import KMeans
from collections import defaultdict
from txtai.embeddings import Embeddings
from yellowbrick.cluster import KElbowVisualizer
from txtai.pipeline import Extractor, Entity, Labels
from transformers import AutoModelForQuestionAnswering, AutoTokenizer, pipeline

from dreamai.core import *
from dreamai.vision import *
from dreamai.imports import *
from dreamai_dl.imports import *


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_parse.ipynb.

# %% auto 0
__all__ = ['col_clusters', 'ColumnCounter', 'get_n_cols', 'pdf_to_cols', 'pdf_cols_to_text', 'pdf_to_text', 'HeadModel',
           'LinBnDrop', 'Flatten', 'AdaptiveConcatPool2d', 'create_head', 'create_model', 'load_cols_model',
           'pdf_to_batch', 'pred_cols']

# %% ../nbs/01_parse.ipynb 3
from .core import *
from .imports import *

# %% ../nbs/01_parse.ipynb 4
def col_clusters(data, data2=None, n_cols=3):
    n_cols = int(n_cols)
    kmeans = KMeans(n_clusters=n_cols, algorithm='elkan', random_state=42).fit(np.reshape(data,(-1,1)))
    idx = np.argsort(kmeans.cluster_centers_.sum(axis=1)).tolist()
    cols = defaultdict(list)
    if data2 is None: data2 = data
    for i,c in enumerate(kmeans.labels_):
        cols[idx.index(c)].append(data2[i])
    return cols

class ColumnCounter(KElbowVisualizer):
    def draw(self):
        pass

def get_n_cols(data, min_c=2, max_c=10, max_n_cols=3):
    model = KMeans(algorithm='elkan', random_state=42)
    visualizer = ColumnCounter(model, k=(min_c, max_c), metric='silhouette')
    visualizer.fit(np.reshape(data,(-1,1)))
    # print(np.reshape(data,(-1,1)))
    if visualizer.elbow_value_ is None:
        return max_n_cols
    return min(visualizer.elbow_value_, max_n_cols)

def pdf_to_cols(data_path, model=None, max_n_cols=3, cols_list=[2,1], device='cpu'):
    pdfs = resolve_data_path(data_path)
    cols_dict = {}
    for file in pdfs:
        if Path(file).suffix == '.pdf':
            try:
                if model is not None:
                    cols_list = pred_cols(file, model, device=device)
                with pdfplumber.open(file) as pdf:
                    pdf_pages = pdf.pages
                    cols_list = cols_list + [None]*(len(pdf_pages)-len(cols_list))
                    pdf_cols = []
                    for page, n_cols in zip(pdf_pages, cols_list):
                        words = page.extract_words(x_tolerance=5)
                        if len(words) == 0:
                            # raise Exception(f'\nCould not extract words from pdf: {str(file)}\nMaybe try extracting tables?')
                            print(f'\nCould not extract words from pdf: {str(file)}. Maybe try extracting tables?')
                            continue
                        word_x = [w['x0'] for w in words]
                        if n_cols is None:
                            try:
                                n_cols = get_n_cols(word_x, max_n_cols=max_n_cols)
                            except:
                                print(f'\nCould not find ideal number of columns for pdf: {str(file)}. Setting to 1.')
                                n_cols = 1
                        # if n_cols == 0:
                            # print(file)
                        cols = col_clusters(word_x, words, n_cols=n_cols)
                        cols = sort_dict({k:sorted(v, key=lambda x: x['top']) for k,v in cols.items()})
                        for k,v in cols.items():
                            paras = []
                            avg_gap = np.mean([w['top']-v[i-1]['top'] for i,w in enumerate(v) if i>0])
                            for i,w in enumerate(v):
                                txt = w['text']
                                if i==0:
                                    paras.append(txt)
                                else:
                                    if w['top']-v[i-1]['bottom'] >= avg_gap:
                                        paras.append(txt)
                                    else:
                                        paras[-1]+=' '+txt.strip()
                            paras = combine_lines(paras)                
                            cols[k] = paras
                        pdf_cols.append(cols)
                    cols = defaultdict(list)
                    for c in pdf_cols:
                        for k,v in c.items():
                            cols[k]+=v
                    cols = sort_dict(cols)
                    cols_dict[str(file)] = cols
            except:
                continue
    return cols_dict

def pdf_cols_to_text(pdf_cols):
    return {k: flatten_list(dict_values(v)) for k,v in pdf_cols.items()}
    # return flatten_list([dict_values(d) for d in dict_values(pdf_cols)])

def pdf_to_text(data_path, model=None, max_n_cols=3, cols_list=[2,1], device='cpu'):
    pdf_cols = pdf_to_cols(data_path, model=model, max_n_cols=max_n_cols, cols_list=cols_list, device=device)
    # return pdf_cols
    return pdf_cols_to_text(pdf_cols)

class HeadModel(nn.Module):
    def __init__(self, pool, linear):
        super().__init__()
        # store_attr(self, 'pool,linear')
        self.pool = pool
        self.linear = linear
    def forward(self, x, meta=None):
        if meta is None:
            return self.linear(self.pool(x))  
        return self.linear(torch.cat([self.pool(x), meta], dim=1))

class LinBnDrop(nn.Sequential):
    "Module grouping `BatchNorm1d`, `Dropout` and `Linear` layers"
    def __init__(self, n_in, n_out, bn=True, p=0., act=None, lin_first=False):
        layers = [nn.BatchNorm1d(n_out if lin_first else n_in)] if bn else []
        if p != 0: layers.append(nn.Dropout(p))
        lin = [nn.Linear(n_in, n_out, bias=not bn)]
        if act is not None: lin.append(act)
        layers = lin+layers if lin_first else layers+lin
        super().__init__(*layers)


class Flatten(nn.Module):
    def forward(self, input):
        return input.view(input.size(0), -1)

class AdaptiveConcatPool2d(nn.Module):
    "Layer that concats `AdaptiveAvgPool2d` and `AdaptiveMaxPool2d`"
    def __init__(self, size=None):
        super().__init__()
        self.size = size or 1
        self.ap = nn.AdaptiveAvgPool2d(self.size)
        self.mp = nn.AdaptiveMaxPool2d(self.size)
    def forward(self, x): return torch.cat([self.mp(x), self.ap(x)], 1)

def create_head(nf, n_out, lin_ftrs=None, ps=0.5, concat_pool=True,
                bn_final=False, lin_first=False, y_range=None, actv=None,
                relu_fn=nn.ReLU(inplace=True), trial=None, num_lin_ftrs=None, n_lin_ftrs=None, trial_num_lin_ftrs=[1,3],
                trial_n_lin_ftrs=[256,512,1024]):
    "Model head that takes `nf` features, runs through `lin_ftrs`, and out `n_out` classes."
    if trial is not None:
        lin_ftrs = [nf]
        num_lin_ftrs = trial.suggest_int("num_lin_ftrs",
                                         trial_num_lin_ftrs[0],
                                         trial_num_lin_ftrs[1])
        for i in range(num_lin_ftrs):
            n_lin_ftrs = trial.suggest_categorical(f"n_lin_ftrs_{i}", trial_n_lin_ftrs)
            # n_lin_ftrs = trial.suggest_categorical(f"n_lin_ftrs_{i}", [1024,1224,1424,1624,1824,2000])
            lin_ftrs.append(n_lin_ftrs)
        lin_ftrs.append(n_out)
    elif num_lin_ftrs is not None and n_lin_ftrs is not None:
        lin_ftrs = [nf]
        for i in range(num_lin_ftrs):
            lin_ftrs.append(n_lin_ftrs[i])
        lin_ftrs.append(n_out)
    else:
        lin_ftrs = [nf, 512, n_out] if lin_ftrs is None else [nf] + lin_ftrs + [n_out]
    ps = [ps]
    if len(ps) == 1: ps = [ps[0]/2] * (len(lin_ftrs)-2) + ps
    actns = [relu_fn] * (len(lin_ftrs)-2) + [None]
    pool = AdaptiveConcatPool2d() if concat_pool else nn.AdaptiveAvgPool2d(1)
    pool_layers = nn.Sequential(*[pool, Flatten()])
    layers = []
    if lin_first: layers.append(nn.Dropout(ps.pop(0)))
    for ni,no,p,actn in zip(lin_ftrs[:-1], lin_ftrs[1:], ps, actns):
        layers += LinBnDrop(ni, no, bn=True, p=p, act=actn, lin_first=lin_first)
    if lin_first: layers.append(nn.Linear(lin_ftrs[-2], n_out))
    if bn_final: layers.append(nn.BatchNorm1d(lin_ftrs[-1], momentum=0.01))
    if actv is not None:
        layers.append(actv)
    # return layers
    layers = nn.Sequential(*layers)
    return HeadModel(pool=pool_layers, linear=layers)

# Initialise the resnet 34 model
def create_model(num_classes=4, lin_ftrs=None, ps=0.5, concat_pool=True, bn_final=False, lin_first=False, actv=None,
                relu_fn=nn.ReLU(inplace=True), trial=None, num_lin_ftrs=None, n_lin_ftrs=None, trial_num_lin_ftrs=[1,3],
                trial_n_lin_ftrs=[256,512,1024]):
        body = nn.Sequential(*list(models.get_model("resnet34",weights="DEFAULT").children())[:-2])
        head_layers = create_head(1024, num_classes, lin_ftrs, ps, concat_pool, bn_final=False, lin_first=False, actv=None,
                relu_fn=nn.ReLU(inplace=True), trial=None, num_lin_ftrs=None, n_lin_ftrs=None, trial_num_lin_ftrs=[1,3],
                trial_n_lin_ftrs=[256,512,1024])
        return nn.Sequential(body,head_layers)
    
def load_cols_model(path, device='cpu'):
    net = create_model( num_classes=3, lin_ftrs=[512,128,64], ps=0.5)
    checkpoint = torch.load(path, map_location=torch.device(device))
    net.load_state_dict(checkpoint, strict=True)
    return net.eval()

def pdf_to_batch(pdf):
    return torch.stack([TF.to_tensor(img) for img in color_fill_pdf_text(pdf)])

def pred_cols(pdf, model, classes=[1,2,3], device='cpu'):
    model = model.eval().to(device)
    batch = pdf_to_batch(pdf)
    batch = batch.to(device)
    with torch.no_grad():
        y_hat = model(batch)
        pred = torch.argmax(y_hat, axis=1) 
    return [classes[p] for p in pred]
